'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _util = require('util');

var _util2 = _interopRequireDefault(_util);

var _v = require('uuid/v4');

var _v2 = _interopRequireDefault(_v);

var _redisApi = require('./redisApi');

var _path = require('path');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

var Z_VALUE_SEPARATOR = '' + String.fromCharCode(0x0) + String.fromCharCode(0x0);
var Z_LEX_MAX_VALUE = String.fromCharCode(0xff);
var TTL_FOR_TEMP_COLLECTION = 60;

var getDocument = function () {
    var _ref = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee(_ref2, collectionName, id) {
        var client = _ref2.client;
        return regeneratorRuntime.wrap(function _callee$(_context) {
            while (1) {
                switch (_context.prev = _context.next) {
                    case 0:
                        _context.next = 2;
                        return (0, _redisApi.hget)(client, collectionName, id);

                    case 2:
                        return _context.abrupt('return', _context.sent);

                    case 3:
                    case 'end':
                        return _context.stop();
                }
            }
        }, _callee, undefined);
    }));

    return function getDocument(_x, _x2, _x3) {
        return _ref.apply(this, arguments);
    };
}();

var validateIndexes = function validateIndexes(indexes, fieldNames, errorMessage) {
    fieldNames.forEach(function (field) {
        if (!indexes[field]) {
            throw new Error(_util2.default.format(errorMessage, field));
        }
    });
};

var validateCriteriaFields = function validateCriteriaFields(indexes, criteria) {
    if (criteria) {
        validateIndexes(indexes, Object.keys(criteria), 'Can\'t find index for \'%s\': you can find by only indexed field');
    }
};

var saveIdsToCollection = function () {
    var _ref3 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2(_ref4, collectionName, ids) {
        var client = _ref4.client;
        var args;
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
            while (1) {
                switch (_context2.prev = _context2.next) {
                    case 0:
                        args = ids.reduce(function (acc, id) {
                            // score
                            acc.push(id);
                            // value
                            acc.push(id);
                            return acc;
                        }, [collectionName]);
                        _context2.next = 3;
                        return (0, _redisApi.zaddMulti)(client, args);

                    case 3:
                        _context2.next = 5;
                        return (0, _redisApi.expire)(client, collectionName, TTL_FOR_TEMP_COLLECTION);

                    case 5:
                    case 'end':
                        return _context2.stop();
                }
            }
        }, _callee2, undefined);
    }));

    return function saveIdsToCollection(_x4, _x5, _x6) {
        return _ref3.apply(this, arguments);
    };
}();

var normalizeIndexValues = function normalizeIndexValues(fieldType, values) {
    return fieldType === 'number' ? values : values.map(function (value) {
        return value.substr(value.lastIndexOf(Z_VALUE_SEPARATOR) + Z_VALUE_SEPARATOR.length);
    });
};

var getIds = function () {
    var _ref5 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee4(repository, collectionName, indexes, criteria) {
        var client, metaCollection, tempPrefix, tempCollectionNames, promises, keys, result;
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
            while (1) {
                switch (_context4.prev = _context4.next) {
                    case 0:
                        client = repository.client, metaCollection = repository.metaCollection;


                        validateCriteriaFields(indexes, criteria);
                        tempPrefix = (0, _v2.default)();
                        tempCollectionNames = {};

                        // todo extract function

                        promises = Object.keys(criteria).map(function () {
                            var _ref6 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee3(fieldName) {
                                var fieldValue, indexName, fieldType, ids;
                                return regeneratorRuntime.wrap(function _callee3$(_context3) {
                                    while (1) {
                                        switch (_context3.prev = _context3.next) {
                                            case 0:
                                                tempCollectionNames[fieldName] = tempPrefix + '_' + fieldName;
                                                fieldValue = criteria[fieldName];
                                                indexName = metaCollection.getFindIndexName(collectionName, fieldName);
                                                fieldType = indexes[fieldName].fieldType;

                                                if (!(fieldType === 'number')) {
                                                    _context3.next = 10;
                                                    break;
                                                }

                                                _context3.next = 7;
                                                return (0, _redisApi.zrangebyscore)(client, indexName, fieldValue, fieldValue);

                                            case 7:
                                                _context3.t0 = _context3.sent;
                                                _context3.next = 13;
                                                break;

                                            case 10:
                                                _context3.next = 12;
                                                return (0, _redisApi.zrangebylex)(client, indexName, '[' + fieldValue + Z_VALUE_SEPARATOR, '(' + fieldValue + Z_VALUE_SEPARATOR + Z_LEX_MAX_VALUE);

                                            case 12:
                                                _context3.t0 = _context3.sent;

                                            case 13:
                                                ids = _context3.t0;

                                                ids = normalizeIndexValues(fieldType, ids);
                                                _context3.next = 17;
                                                return saveIdsToCollection(repository, tempCollectionNames[fieldName], ids);

                                            case 17:
                                                return _context3.abrupt('return', _context3.sent);

                                            case 18:
                                            case 'end':
                                                return _context3.stop();
                                        }
                                    }
                                }, _callee3, undefined);
                            }));

                            return function (_x11) {
                                return _ref6.apply(this, arguments);
                            };
                        }());
                        _context4.next = 7;
                        return Promise.all(promises);

                    case 7:
                        keys = Object.values(tempCollectionNames);
                        _context4.next = 10;
                        return (0, _redisApi.zinterstore)(client, tempPrefix, keys.length, keys);

                    case 10:
                        _context4.next = 12;
                        return (0, _redisApi.expire)(client, tempPrefix, TTL_FOR_TEMP_COLLECTION);

                    case 12:
                        _context4.next = 14;
                        return (0, _redisApi.zrange)(client, tempPrefix);

                    case 14:
                        result = _context4.sent;
                        _context4.next = 17;
                        return _redisApi.del.apply(undefined, [client].concat(_toConsumableArray(keys.concat(tempPrefix))));

                    case 17:
                        return _context4.abrupt('return', result);

                    case 18:
                    case 'end':
                        return _context4.stop();
                }
            }
        }, _callee4, undefined);
    }));

    return function getIds(_x7, _x8, _x9, _x10) {
        return _ref5.apply(this, arguments);
    };
}();

var populateNumberIndex = function () {
    var _ref7 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee5(client, collectionName, score, id) {
        return regeneratorRuntime.wrap(function _callee5$(_context5) {
            while (1) {
                switch (_context5.prev = _context5.next) {
                    case 0:
                        _context5.next = 2;
                        return (0, _redisApi.zadd)(client, collectionName, score, id);

                    case 2:
                    case 'end':
                        return _context5.stop();
                }
            }
        }, _callee5, undefined);
    }));

    return function populateNumberIndex(_x12, _x13, _x14, _x15) {
        return _ref7.apply(this, arguments);
    };
}();

var populateStringIndex = function () {
    var _ref8 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee6(client, collectionName, value, id) {
        return regeneratorRuntime.wrap(function _callee6$(_context6) {
            while (1) {
                switch (_context6.prev = _context6.next) {
                    case 0:
                        _context6.next = 2;
                        return (0, _redisApi.zadd)(client, collectionName, 0, '' + value + Z_VALUE_SEPARATOR + id);

                    case 2:
                    case 'end':
                        return _context6.stop();
                }
            }
        }, _callee6, undefined);
    }));

    return function populateStringIndex(_x16, _x17, _x18, _x19) {
        return _ref8.apply(this, arguments);
    };
}();

var populateFindIndex = function () {
    var _ref9 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee7(client, collectionName, fieldName, fieldType, indexName, id, value) {
        return regeneratorRuntime.wrap(function _callee7$(_context7) {
            while (1) {
                switch (_context7.prev = _context7.next) {
                    case 0:
                        if (!(fieldType === 'number')) {
                            _context7.next = 7;
                            break;
                        }

                        if (!(typeof value !== 'number')) {
                            _context7.next = 3;
                            break;
                        }

                        throw new Error('Can\'t update index \'' + fieldName + '\' value: Value \'' + value + '\' is not number.');

                    case 3:
                        _context7.next = 5;
                        return populateNumberIndex(client, indexName, value, id);

                    case 5:
                        _context7.next = 9;
                        break;

                    case 7:
                        _context7.next = 9;
                        return populateStringIndex(client, indexName, value, id);

                    case 9:
                    case 'end':
                        return _context7.stop();
                }
            }
        }, _callee7, undefined);
    }));

    return function populateFindIndex(_x20, _x21, _x22, _x23, _x24, _x25, _x26) {
        return _ref9.apply(this, arguments);
    };
}();

var populateSortIndex = function () {
    var _ref10 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee8(client, collectionName, value, id) {
        return regeneratorRuntime.wrap(function _callee8$(_context8) {
            while (1) {
                switch (_context8.prev = _context8.next) {
                    case 0:
                        _context8.next = 2;
                        return (0, _redisApi.hset)(client, collectionName, id, value);

                    case 2:
                    case 'end':
                        return _context8.stop();
                }
            }
        }, _callee8, undefined);
    }));

    return function populateSortIndex(_x27, _x28, _x29, _x30) {
        return _ref10.apply(this, arguments);
    };
}();

var updateIndex = function () {
    var _ref11 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee9(_ref12, collectionName, id, document, _ref13) {
        var client = _ref12.client,
            metaCollection = _ref12.metaCollection;
        var fieldName = _ref13.fieldName,
            fieldType = _ref13.fieldType;
        var value, findIndexName, sortIndexName;
        return regeneratorRuntime.wrap(function _callee9$(_context9) {
            while (1) {
                switch (_context9.prev = _context9.next) {
                    case 0:
                        value = document[fieldName];
                        findIndexName = metaCollection.getFindIndexName(collectionName, fieldName);
                        _context9.next = 4;
                        return populateFindIndex(client, collectionName, fieldName, fieldType, findIndexName, id, value);

                    case 4:
                        sortIndexName = metaCollection.getSortIndexName(collectionName, fieldName);
                        _context9.next = 7;
                        return populateSortIndex(client, sortIndexName, value, id);

                    case 7:
                    case 'end':
                        return _context9.stop();
                }
            }
        }, _callee9, undefined);
    }));

    return function updateIndex(_x31, _x32, _x33, _x34, _x35) {
        return _ref11.apply(this, arguments);
    };
}();

var updateIndexes = function () {
    var _ref14 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee11(repository, collectionName, id, document) {
        var metaCollection, indexes, promises;
        return regeneratorRuntime.wrap(function _callee11$(_context11) {
            while (1) {
                switch (_context11.prev = _context11.next) {
                    case 0:
                        metaCollection = repository.metaCollection;
                        _context11.next = 3;
                        return metaCollection.getIndexes(collectionName);

                    case 3:
                        indexes = _context11.sent;
                        promises = Object.keys(indexes).map(function () {
                            var _ref15 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee10(fieldName) {
                                var index;
                                return regeneratorRuntime.wrap(function _callee10$(_context10) {
                                    while (1) {
                                        switch (_context10.prev = _context10.next) {
                                            case 0:
                                                index = indexes[fieldName];
                                                _context10.next = 3;
                                                return updateIndex(repository, collectionName, id, document, index);

                                            case 3:
                                            case 'end':
                                                return _context10.stop();
                                        }
                                    }
                                }, _callee10, undefined);
                            }));

                            return function (_x40) {
                                return _ref15.apply(this, arguments);
                            };
                        }());
                        _context11.next = 7;
                        return Promise.all(promises);

                    case 7:
                    case 'end':
                        return _context11.stop();
                }
            }
        }, _callee11, undefined);
    }));

    return function updateIndexes(_x36, _x37, _x38, _x39) {
        return _ref14.apply(this, arguments);
    };
}();

var criteriaIsEmpty = function criteriaIsEmpty(criteria) {
    return !(criteria && Object.keys(criteria).length);
};

var count = function () {
    var _ref16 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee12(_ref17, collectionName, criteria) {
        var client = _ref17.client;
        return regeneratorRuntime.wrap(function _callee12$(_context12) {
            while (1) {
                switch (_context12.prev = _context12.next) {
                    case 0:
                        _context12.next = 2;
                        return (0, _redisApi.hlen)(client, collectionName);

                    case 2:
                        return _context12.abrupt('return', _context12.sent);

                    case 3:
                    case 'end':
                        return _context12.stop();
                }
            }
        }, _callee12, undefined);
    }));

    return function count(_x41, _x42, _x43) {
        return _ref16.apply(this, arguments);
    };
}();

var validateSortFields = function validateSortFields(indexes, sort) {
    if (sort) {
        var fieldNames = Object.keys(sort);
        if (fieldNames.length > 1) {
            throw new Error('You can sort by only one field');
        }
        validateIndexes(indexes, fieldNames, 'Can\'t find index for \'%s\': you can sort by only indexed field');
    }
};

var find = function () {
    var _ref18 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee15(repository, collectionName, criteria, _ref19) {
        var _ref19$skip = _ref19.skip,
            skip = _ref19$skip === undefined ? 0 : _ref19$skip,
            _ref19$limit = _ref19.limit,
            limit = _ref19$limit === undefined ? -1 : _ref19$limit,
            _ref19$sort = _ref19.sort,
            sort = _ref19$sort === undefined ? null : _ref19$sort;

        var client, metaCollection, indexes, ids, indexName, sortField, order, fieldType, sortIndexName, tempSortTable, _promises, stop, end, promises;

        return regeneratorRuntime.wrap(function _callee15$(_context15) {
            while (1) {
                switch (_context15.prev = _context15.next) {
                    case 0:
                        client = repository.client, metaCollection = repository.metaCollection;
                        _context15.next = 3;
                        return metaCollection.getIndexes(collectionName);

                    case 3:
                        indexes = _context15.sent;

                        validateCriteriaFields(indexes, criteria);
                        validateSortFields(indexes, sort);
                        ids = void 0;

                        if (!criteriaIsEmpty(criteria)) {
                            _context15.next = 14;
                            break;
                        }

                        indexName = metaCollection.getFindIndexName(collectionName, '_id');
                        _context15.next = 11;
                        return (0, _redisApi.zrange)(client, indexName);

                    case 11:
                        ids = _context15.sent;
                        _context15.next = 17;
                        break;

                    case 14:
                        _context15.next = 16;
                        return getIds(repository, collectionName, indexes, criteria);

                    case 16:
                        ids = _context15.sent;

                    case 17:
                        if (!sort) {
                            _context15.next = 43;
                            break;
                        }

                        sortField = Object.keys(sort)[0];
                        order = sort[sortField];
                        fieldType = indexes[sortField].fieldType;
                        sortIndexName = metaCollection.getSortIndexName(collectionName, sortField);
                        tempSortTable = (0, _v2.default)();
                        _promises = ids.map(function () {
                            var _ref20 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee13(id) {
                                var value;
                                return regeneratorRuntime.wrap(function _callee13$(_context13) {
                                    while (1) {
                                        switch (_context13.prev = _context13.next) {
                                            case 0:
                                                _context13.next = 2;
                                                return (0, _redisApi.hget)(client, sortIndexName, id);

                                            case 2:
                                                value = _context13.sent;
                                                _context13.next = 5;
                                                return populateFindIndex(client, collectionName, sortField, fieldType, tempSortTable, id, value);

                                            case 5:
                                            case 'end':
                                                return _context13.stop();
                                        }
                                    }
                                }, _callee13, undefined);
                            }));

                            return function (_x48) {
                                return _ref20.apply(this, arguments);
                            };
                        }());
                        _context15.next = 26;
                        return Promise.all(_promises);

                    case 26:
                        _context15.next = 28;
                        return (0, _redisApi.expire)(client, tempSortTable, TTL_FOR_TEMP_COLLECTION);

                    case 28:
                        stop = limit === -1 ? -1 : skip + limit - 1;

                        if (!(order === 1)) {
                            _context15.next = 35;
                            break;
                        }

                        _context15.next = 32;
                        return (0, _redisApi.zrange)(client, tempSortTable, skip, stop);

                    case 32:
                        _context15.t0 = _context15.sent;
                        _context15.next = 38;
                        break;

                    case 35:
                        _context15.next = 37;
                        return (0, _redisApi.zrevrange)(client, tempSortTable, skip, stop);

                    case 37:
                        _context15.t0 = _context15.sent;

                    case 38:
                        ids = _context15.t0;


                        (0, _redisApi.del)(client, tempSortTable);
                        ids = normalizeIndexValues(fieldType, ids);
                        _context15.next = 45;
                        break;

                    case 43:
                        end = limit === -1 ? undefined : skip + limit;

                        ids = ids.slice(skip, end);

                    case 45:
                        promises = ids.map(function () {
                            var _ref21 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee14(id) {
                                return regeneratorRuntime.wrap(function _callee14$(_context14) {
                                    while (1) {
                                        switch (_context14.prev = _context14.next) {
                                            case 0:
                                                _context14.next = 2;
                                                return getDocument(repository, collectionName, id);

                                            case 2:
                                                return _context14.abrupt('return', _context14.sent);

                                            case 3:
                                            case 'end':
                                                return _context14.stop();
                                        }
                                    }
                                }, _callee14, undefined);
                            }));

                            return function (_x49) {
                                return _ref21.apply(this, arguments);
                            };
                        }());
                        _context15.next = 48;
                        return Promise.all(promises);

                    case 48:
                        return _context15.abrupt('return', _context15.sent);

                    case 49:
                    case 'end':
                        return _context15.stop();
                }
            }
        }, _callee15, undefined);
    }));

    return function find(_x44, _x45, _x46, _x47) {
        return _ref18.apply(this, arguments);
    };
}();

var findOne = function () {
    var _ref22 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee16(repository, collectionName, criteria, _ref23) {
        var _ref23$skip = _ref23.skip,
            skip = _ref23$skip === undefined ? 0 : _ref23$skip,
            _ref23$order = _ref23.order,
            order = _ref23$order === undefined ? 0 : _ref23$order;
        var limit;
        return regeneratorRuntime.wrap(function _callee16$(_context16) {
            while (1) {
                switch (_context16.prev = _context16.next) {
                    case 0:
                        limit = 1;
                        return _context16.abrupt('return', find(repository, collectionName, criteria, { skip: skip, limit: limit, order: order }));

                    case 2:
                    case 'end':
                        return _context16.stop();
                }
            }
        }, _callee16, undefined);
    }));

    return function findOne(_x50, _x51, _x52, _x53) {
        return _ref22.apply(this, arguments);
    };
}();

function wrapFind(repository, initialFind, collectionName, expression) {
    var resultPromise = Promise.resolve();
    var requestChain = {
        // skip: null,
        // limit: null,
        sort: null
    };
    var sanitizeError = ''; //sanitizeSearchExpression(expression);

    ['skip', 'limit', 'sort'].forEach(function (cmd) {
        resultPromise[cmd] = function (value) {
            requestChain[cmd] = value;
            return resultPromise;
        };
    });

    var originalThen = resultPromise.then.bind(resultPromise);
    var boundExecFind = (!sanitizeError ? initialFind : Promise.reject.bind(Promise, new Error(sanitizeError))).bind(null, repository, collectionName, expression, requestChain);

    resultPromise.then = function () {
        var _originalThen;

        return (_originalThen = originalThen(boundExecFind)).then.apply(_originalThen, arguments);
    };

    resultPromise.catch = function () {
        var _originalThen2;

        return (_originalThen2 = originalThen(boundExecFind)).catch.apply(_originalThen2, arguments);
    };

    return resultPromise;
}

var insert = function () {
    var _ref24 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee17(repository, collectionName, document) {
        var client, metaCollection, _id, member;

        return regeneratorRuntime.wrap(function _callee17$(_context17) {
            while (1) {
                switch (_context17.prev = _context17.next) {
                    case 0:
                        client = repository.client, metaCollection = repository.metaCollection;
                        _context17.next = 3;
                        return metaCollection.getNextId(collectionName);

                    case 3:
                        _id = _context17.sent;
                        member = _extends({ _id: _id }, document);
                        _context17.next = 7;
                        return (0, _redisApi.hset)(client, collectionName, _id, member);

                    case 7:
                        _context17.next = 9;
                        return updateIndexes(repository, collectionName, _id, member);

                    case 9:
                    case 'end':
                        return _context17.stop();
                }
            }
        }, _callee17, undefined);
    }));

    return function insert(_x54, _x55, _x56) {
        return _ref24.apply(this, arguments);
    };
}();

var removeAll = function () {
    var _ref25 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee19(_ref26, collectionName) {
        var client = _ref26.client,
            metaCollection = _ref26.metaCollection;
        var indexes;
        return regeneratorRuntime.wrap(function _callee19$(_context19) {
            while (1) {
                switch (_context19.prev = _context19.next) {
                    case 0:
                        _context19.next = 2;
                        return metaCollection.getIndexes(collectionName);

                    case 2:
                        indexes = _context19.sent;

                        Object.keys(indexes).forEach(function () {
                            var _ref27 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee18(key) {
                                var fieldName, findIndexName, sortIndexName;
                                return regeneratorRuntime.wrap(function _callee18$(_context18) {
                                    while (1) {
                                        switch (_context18.prev = _context18.next) {
                                            case 0:
                                                fieldName = indexes[key].fieldName;
                                                findIndexName = metaCollection.getFindIndexName(collectionName, fieldName);
                                                _context18.next = 4;
                                                return (0, _redisApi.del)(client, findIndexName);

                                            case 4:
                                                sortIndexName = metaCollection.getSortIndexName(collectionName, fieldName);
                                                _context18.next = 7;
                                                return (0, _redisApi.del)(client, sortIndexName);

                                            case 7:
                                            case 'end':
                                                return _context18.stop();
                                        }
                                    }
                                }, _callee18, undefined);
                            }));

                            return function (_x59) {
                                return _ref27.apply(this, arguments);
                            };
                        }());
                        _context19.next = 6;
                        return (0, _redisApi.del)(client, collectionName);

                    case 6:
                    case 'end':
                        return _context19.stop();
                }
            }
        }, _callee19, undefined);
    }));

    return function removeAll(_x57, _x58) {
        return _ref25.apply(this, arguments);
    };
}();

var removeIdFromIndex = function () {
    var _ref28 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee20(_ref29, collectionName, id, document, index) {
        var client = _ref29.client,
            metaCollection = _ref29.metaCollection;
        var fieldName, fieldType, findIndexName, fieldValue;
        return regeneratorRuntime.wrap(function _callee20$(_context20) {
            while (1) {
                switch (_context20.prev = _context20.next) {
                    case 0:
                        fieldName = index.fieldName, fieldType = index.fieldType;
                        findIndexName = metaCollection.getFindIndexName(collectionName, fieldName);

                        if (!(fieldType === 'number')) {
                            _context20.next = 7;
                            break;
                        }

                        _context20.next = 5;
                        return (0, _redisApi.zrem)(client, findIndexName, id);

                    case 5:
                        _context20.next = 10;
                        break;

                    case 7:
                        fieldValue = document[fieldName];
                        _context20.next = 10;
                        return (0, _redisApi.zremrangebylex)(client, findIndexName, '[' + fieldValue + Z_VALUE_SEPARATOR + id, '(' + fieldValue + Z_VALUE_SEPARATOR + id);

                    case 10:
                        _context20.next = 12;
                        return (0, _redisApi.hdel)(client, metaCollection.getSortIndexName(collectionName, fieldName));

                    case 12:
                    case 'end':
                        return _context20.stop();
                }
            }
        }, _callee20, undefined);
    }));

    return function removeIdFromIndex(_x60, _x61, _x62, _x63, _x64) {
        return _ref28.apply(this, arguments);
    };
}();

var removeIdsFromIndexes = function () {
    var _ref30 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee23(repository, collectionName, indexes, ids) {
        var promises;
        return regeneratorRuntime.wrap(function _callee23$(_context23) {
            while (1) {
                switch (_context23.prev = _context23.next) {
                    case 0:
                        promises = ids.map(function () {
                            var _ref31 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee22(id) {
                                var document, proms;
                                return regeneratorRuntime.wrap(function _callee22$(_context22) {
                                    while (1) {
                                        switch (_context22.prev = _context22.next) {
                                            case 0:
                                                _context22.next = 2;
                                                return getDocument(repository, collectionName, id);

                                            case 2:
                                                document = _context22.sent;
                                                proms = Object.keys(indexes).map(function () {
                                                    var _ref32 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee21(key) {
                                                        return regeneratorRuntime.wrap(function _callee21$(_context21) {
                                                            while (1) {
                                                                switch (_context21.prev = _context21.next) {
                                                                    case 0:
                                                                        _context21.next = 2;
                                                                        return removeIdFromIndex(repository, collectionName, id, document, indexes[key]);

                                                                    case 2:
                                                                        return _context21.abrupt('return', _context21.sent);

                                                                    case 3:
                                                                    case 'end':
                                                                        return _context21.stop();
                                                                }
                                                            }
                                                        }, _callee21, undefined);
                                                    }));

                                                    return function (_x70) {
                                                        return _ref32.apply(this, arguments);
                                                    };
                                                }());
                                                _context22.next = 6;
                                                return Promise.all(proms);

                                            case 6:
                                            case 'end':
                                                return _context22.stop();
                                        }
                                    }
                                }, _callee22, undefined);
                            }));

                            return function (_x69) {
                                return _ref31.apply(this, arguments);
                            };
                        }());
                        _context23.next = 3;
                        return Promise.all(promises);

                    case 3:
                    case 'end':
                        return _context23.stop();
                }
            }
        }, _callee23, undefined);
    }));

    return function removeIdsFromIndexes(_x65, _x66, _x67, _x68) {
        return _ref30.apply(this, arguments);
    };
}();

var removeIds = function () {
    var _ref33 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee24(repository, collectionName, indexes, ids) {
        return regeneratorRuntime.wrap(function _callee24$(_context24) {
            while (1) {
                switch (_context24.prev = _context24.next) {
                    case 0:
                        _context24.next = 2;
                        return removeIdsFromIndexes(repository, collectionName, indexes, ids);

                    case 2:
                        _context24.next = 4;
                        return _redisApi.hdel.apply(undefined, [repository.client, collectionName].concat(_toConsumableArray(ids)));

                    case 4:
                    case 'end':
                        return _context24.stop();
                }
            }
        }, _callee24, undefined);
    }));

    return function removeIds(_x71, _x72, _x73, _x74) {
        return _ref33.apply(this, arguments);
    };
}();

var remove = function () {
    var _ref34 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee25(repository, collectionName, criteria) {
        var metaCollection, indexes, ids;
        return regeneratorRuntime.wrap(function _callee25$(_context25) {
            while (1) {
                switch (_context25.prev = _context25.next) {
                    case 0:
                        metaCollection = repository.metaCollection;

                        if (!(!criteria || !Object.keys(criteria).length)) {
                            _context25.next = 5;
                            break;
                        }

                        _context25.next = 4;
                        return removeAll(repository, collectionName);

                    case 4:
                        return _context25.abrupt('return', _context25.sent);

                    case 5:
                        _context25.next = 7;
                        return metaCollection.getIndexes(collectionName);

                    case 7:
                        indexes = _context25.sent;
                        _context25.next = 10;
                        return getIds(repository, collectionName, indexes, criteria);

                    case 10:
                        ids = _context25.sent;
                        _context25.next = 13;
                        return removeIds(repository, collectionName, indexes, ids);

                    case 13:
                    case 'end':
                        return _context25.stop();
                }
            }
        }, _callee25, undefined);
    }));

    return function remove(_x75, _x76, _x77) {
        return _ref34.apply(this, arguments);
    };
}();

var applyDotNotation = function () {
    var _ref35 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee26(document, fieldName, fieldOptions, cb) {
        var fields;
        return regeneratorRuntime.wrap(function _callee26$(_context26) {
            while (1) {
                switch (_context26.prev = _context26.next) {
                    case 0:
                        fields = fieldName.split('.');

                        if (fields.length > 1) {
                            fields.reduce(function (acc, field, currentIndex) {
                                if (fields.length === currentIndex - 1) {
                                    cb(acc, fieldName, fieldOptions, false);
                                }
                                return acc[fieldName];
                            }, document);
                        } else {
                            cb(document, fieldName, fieldOptions, true);
                        }

                    case 2:
                    case 'end':
                        return _context26.stop();
                }
            }
        }, _callee26, undefined);
    }));

    return function applyDotNotation(_x78, _x79, _x80, _x81) {
        return _ref35.apply(this, arguments);
    };
}();

var applyFields = function () {
    var _ref36 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee28(document, options, cb) {
        var promises;
        return regeneratorRuntime.wrap(function _callee28$(_context28) {
            while (1) {
                switch (_context28.prev = _context28.next) {
                    case 0:
                        promises = Object.keys(options).map(function () {
                            var _ref37 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee27(fieldName) {
                                return regeneratorRuntime.wrap(function _callee27$(_context27) {
                                    while (1) {
                                        switch (_context27.prev = _context27.next) {
                                            case 0:
                                                return _context27.abrupt('return', applyDotNotation(document, fieldName, options[fieldName], cb));

                                            case 1:
                                            case 'end':
                                                return _context27.stop();
                                        }
                                    }
                                }, _callee27, undefined);
                            }));

                            return function (_x85) {
                                return _ref37.apply(this, arguments);
                            };
                        }());
                        _context28.next = 3;
                        return Promise.all(promises);

                    case 3:
                    case 'end':
                        return _context28.stop();
                }
            }
        }, _callee28, undefined);
    }));

    return function applyFields(_x82, _x83, _x84) {
        return _ref36.apply(this, arguments);
    };
}();

var updateHandlers = {
    $unset: function () {
        var _ref38 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee29(repository, collectionName, indexes, id, partDoc, fieldName, isRootLevel) {
            var index;
            return regeneratorRuntime.wrap(function _callee29$(_context29) {
                while (1) {
                    switch (_context29.prev = _context29.next) {
                        case 0:
                            if (!isRootLevel) {
                                _context29.next = 5;
                                break;
                            }

                            index = indexes[fieldName];

                            if (!index) {
                                _context29.next = 5;
                                break;
                            }

                            _context29.next = 5;
                            return removeIdFromIndex(repository, collectionName, id, partDoc, index);

                        case 5:

                            if (Array.isArray(partDoc[fieldName])) {
                                partDoc[fieldName] = null;
                            } else {
                                delete partDoc[fieldName];
                            }

                        case 6:
                        case 'end':
                            return _context29.stop();
                    }
                }
            }, _callee29, undefined);
        }));

        return function $unset(_x86, _x87, _x88, _x89, _x90, _x91, _x92) {
            return _ref38.apply(this, arguments);
        };
    }(),
    $inc: function () {
        var _ref39 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee30(repository, collectionName, indexes, id, partDoc, fieldName, isRootLevel, fieldOptions) {
            var index, _index;

            return regeneratorRuntime.wrap(function _callee30$(_context30) {
                while (1) {
                    switch (_context30.prev = _context30.next) {
                        case 0:
                            if (!isRootLevel) {
                                _context30.next = 5;
                                break;
                            }

                            index = indexes[fieldName];

                            if (!index) {
                                _context30.next = 5;
                                break;
                            }

                            _context30.next = 5;
                            return removeIdFromIndex(repository, collectionName, id, partDoc, index);

                        case 5:

                            partDoc[fieldName] += fieldOptions;

                            if (!isRootLevel) {
                                _context30.next = 11;
                                break;
                            }

                            _index = indexes[fieldName];

                            if (!_index) {
                                _context30.next = 11;
                                break;
                            }

                            _context30.next = 11;
                            return updateIndex(repository, collectionName, id, document, _index);

                        case 11:
                        case 'end':
                            return _context30.stop();
                    }
                }
            }, _callee30, undefined);
        }));

        return function $inc(_x93, _x94, _x95, _x96, _x97, _x98, _x99, _x100) {
            return _ref39.apply(this, arguments);
        };
    }(),
    $push: function () {
        var _ref40 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee31(repository, collectionName, indexes, id, partDoc, fieldName, isRootLevel, fieldOptions) {
            var index, _index2;

            return regeneratorRuntime.wrap(function _callee31$(_context31) {
                while (1) {
                    switch (_context31.prev = _context31.next) {
                        case 0:
                            if (Array.isArray(partDoc[fieldName])) {
                                _context31.next = 2;
                                break;
                            }

                            throw new Error('The field \'' + fieldName + '\' is not array in document { _id: ' + id + ' }');

                        case 2:
                            if (!isRootLevel) {
                                _context31.next = 7;
                                break;
                            }

                            index = indexes[fieldName];

                            if (!index) {
                                _context31.next = 7;
                                break;
                            }

                            _context31.next = 7;
                            return removeIdFromIndex(repository, collectionName, id, partDoc, index);

                        case 7:
                            document[fieldName] = fieldOptions;

                            if (!isRootLevel) {
                                _context31.next = 13;
                                break;
                            }

                            _index2 = indexes[fieldName];

                            if (!_index2) {
                                _context31.next = 13;
                                break;
                            }

                            _context31.next = 13;
                            return updateIndex(repository, collectionName, id, document, _index2);

                        case 13:
                        case 'end':
                            return _context31.stop();
                    }
                }
            }, _callee31, undefined);
        }));

        return function $push(_x101, _x102, _x103, _x104, _x105, _x106, _x107, _x108) {
            return _ref40.apply(this, arguments);
        };
    }(),
    $pull: function () {
        var _ref41 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee32(repository, collectionName, indexes, id, document, options) {
            return regeneratorRuntime.wrap(function _callee32$(_context32) {
                while (1) {
                    switch (_context32.prev = _context32.next) {
                        case 0:
                            throw new Error('Not implemented!!!');

                        case 1:
                        case 'end':
                            return _context32.stop();
                    }
                }
            }, _callee32, undefined);
        }));

        return function $pull(_x109, _x110, _x111, _x112, _x113, _x114) {
            return _ref41.apply(this, arguments);
        };
    }()
};

var update = function () {
    var _ref42 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee35(repository, collectionName, criteria, operators) {
        var client, metaCollection, indexes, ids;
        return regeneratorRuntime.wrap(function _callee35$(_context35) {
            while (1) {
                switch (_context35.prev = _context35.next) {
                    case 0:
                        client = repository.client, metaCollection = repository.metaCollection;
                        _context35.next = 3;
                        return metaCollection.getIndexes(collectionName);

                    case 3:
                        indexes = _context35.sent;
                        _context35.next = 6;
                        return getIds(repository, collectionName, indexes, criteria);

                    case 6:
                        ids = _context35.sent;

                        ids.forEach(function () {
                            var _ref43 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee34(id) {
                                var document, promises;
                                return regeneratorRuntime.wrap(function _callee34$(_context34) {
                                    while (1) {
                                        switch (_context34.prev = _context34.next) {
                                            case 0:
                                                _context34.next = 2;
                                                return getDocument(repository, collectionName, id);

                                            case 2:
                                                document = _context34.sent;
                                                promises = Object.keys(operators).map(function () {
                                                    var _ref44 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee33(operatorName) {
                                                        var handler, operatorOptions;
                                                        return regeneratorRuntime.wrap(function _callee33$(_context33) {
                                                            while (1) {
                                                                switch (_context33.prev = _context33.next) {
                                                                    case 0:
                                                                        handler = updateHandlers[operatorName];

                                                                        if (handler) {
                                                                            _context33.next = 3;
                                                                            break;
                                                                        }

                                                                        throw new Error('Invalid update operator');

                                                                    case 3:
                                                                        operatorOptions = operators[operatorName];
                                                                        _context33.next = 6;
                                                                        return applyFields(document, operatorOptions, function (doc, fieldName, fieldOptions, isRootLevel) {
                                                                            handler(repository, collectionName, indexes, id, doc, fieldName, isRootLevel, fieldOptions);
                                                                        });

                                                                    case 6:
                                                                    case 'end':
                                                                        return _context33.stop();
                                                                }
                                                            }
                                                        }, _callee33, undefined);
                                                    }));

                                                    return function (_x120) {
                                                        return _ref44.apply(this, arguments);
                                                    };
                                                }());
                                                _context34.next = 6;
                                                return Promise.all(promises);

                                            case 6:
                                                _context34.next = 8;
                                                return (0, _redisApi.hset)(client, collectionName, id, document);

                                            case 8:
                                            case 'end':
                                                return _context34.stop();
                                        }
                                    }
                                }, _callee34, undefined);
                            }));

                            return function (_x119) {
                                return _ref43.apply(this, arguments);
                            };
                        }());

                    case 8:
                    case 'end':
                        return _context35.stop();
                }
            }
        }, _callee35, undefined);
    }));

    return function update(_x115, _x116, _x117, _x118) {
        return _ref42.apply(this, arguments);
    };
}();

var createIndex = function () {
    var _ref45 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee36(repository, collectionName, _ref46) {
        var fieldName = _ref46.fieldName,
            fieldType = _ref46.fieldType,
            _ref46$order = _ref46.order,
            order = _ref46$order === undefined ? 1 : _ref46$order;
        var metaCollection;
        return regeneratorRuntime.wrap(function _callee36$(_context36) {
            while (1) {
                switch (_context36.prev = _context36.next) {
                    case 0:
                        metaCollection = repository.metaCollection;

                        if (fieldName) {
                            _context36.next = 3;
                            break;
                        }

                        throw new Error('`ensureIndex` - invalid fieldName');

                    case 3:
                        if (fieldType === 'string' || fieldType === 'number') {
                            _context36.next = 5;
                            break;
                        }

                        throw new Error('`ensureIndex` - invalid fieldType');

                    case 5:
                        if (order === 1 || order === -1) {
                            _context36.next = 7;
                            break;
                        }

                        throw new Error('`ensureIndex` - invalid order type: you can use only 1 or -1');

                    case 7:
                        _context36.next = 9;
                        return count(repository, collectionName, {});

                    case 9:
                        _context36.t0 = _context36.sent;

                        if (!(_context36.t0 > 0)) {
                            _context36.next = 12;
                            break;
                        }

                        throw new Error('Can\'t create the index by \'' + fieldName + '\' field ' + ('when collection \'' + collectionName + '\' have documents'));

                    case 12:
                        _context36.next = 14;
                        return metaCollection.ensureIndex(collectionName, { fieldName: fieldName, fieldType: fieldType, order: order });

                    case 14:
                    case 'end':
                        return _context36.stop();
                }
            }
        }, _callee36, undefined);
    }));

    return function createIndex(_x121, _x122, _x123) {
        return _ref45.apply(this, arguments);
    };
}();

var ensureIndex = function () {
    var _ref47 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee37(repository, collectionName, options) {
        var metaCollection, fieldName, fieldType, _options$order, order, indexes, idx;

        return regeneratorRuntime.wrap(function _callee37$(_context37) {
            while (1) {
                switch (_context37.prev = _context37.next) {
                    case 0:
                        metaCollection = repository.metaCollection;
                        fieldName = options.fieldName, fieldType = options.fieldType, _options$order = options.order, order = _options$order === undefined ? 1 : _options$order;
                        indexes = metaCollection.getIndexes(collectionName);

                        if (!(indexes && indexes[fieldName])) {
                            _context37.next = 7;
                            break;
                        }

                        idx = indexes[fieldName];

                        if (!(idx.fieldName !== fieldName || idx.fieldType !== fieldType || idx.order !== order)) {
                            _context37.next = 7;
                            break;
                        }

                        throw new Error('Collection \'' + collectionName + '\' has index by \'' + fieldName + '\' field ' + 'but fieldType or order is different');

                    case 7:
                        return _context37.abrupt('return', createIndex(repository, collectionName, options));

                    case 8:
                    case 'end':
                        return _context37.stop();
                }
            }
        }, _callee37, undefined);
    }));

    return function ensureIndex(_x124, _x125, _x126) {
        return _ref47.apply(this, arguments);
    };
}();

var removeIndex = function () {
    var _ref48 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee38(_ref49, collectionName, fieldName) {
        var client = _ref49.client,
            metaCollection = _ref49.metaCollection;
        return regeneratorRuntime.wrap(function _callee38$(_context38) {
            while (1) {
                switch (_context38.prev = _context38.next) {
                    case 0:
                        _context38.next = 2;
                        return metaCollection.removeIndex(collectionName, fieldName);

                    case 2:
                        _context38.next = 4;
                        return (0, _redisApi.del)(client, metaCollection.getFindIndexName(collectionName, fieldName));

                    case 4:
                        _context38.next = 6;
                        return (0, _redisApi.del)(client, metaCollection.getSortIndexName(collectionName, fieldName));

                    case 6:
                    case 'end':
                        return _context38.stop();
                }
            }
        }, _callee38, undefined);
    }));

    return function removeIndex(_x127, _x128, _x129) {
        return _ref48.apply(this, arguments);
    };
}();

var collection = function collection(repository, collectionName) {
    return Object.freeze({
        count: count.bind(null, repository, collectionName),
        find: wrapFind.bind(null, repository, find, collectionName),
        findOne: wrapFind.bind(null, repository, findOne, collectionName),
        insert: insert.bind(null, repository, collectionName),
        remove: remove.bind(null, repository, collectionName),
        update: update.bind(null, repository, collectionName),

        ensureIndex: ensureIndex.bind(null, repository, collectionName),
        removeIndex: removeIndex.bind(null, repository, collectionName)

        // promise
        // then: array => {},
        // catch: () => {}
    });
};

exports.default = collection;
module.exports = exports['default'];