'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var initProjection = function () {
    var _ref = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee(repository) {
        return regeneratorRuntime.wrap(function _callee$(_context) {
            while (1) {
                switch (_context.prev = _context.next) {
                    case 0:
                        _context.next = 2;
                        return repository.connectionPromise;

                    case 2:
                        if (!(repository.lastTimestamp !== 0)) {
                            _context.next = 4;
                            break;
                        }

                        return _context.abrupt('return');

                    case 4:
                        repository.lastTimestamp = 1;

                        _context.prev = 5;
                        _context.next = 8;
                        return repository.initHandler(repository.writeInterface);

                    case 8:
                        _context.next = 13;
                        break;

                    case 10:
                        _context.prev = 10;
                        _context.t0 = _context['catch'](5);

                        repository.internalError = _context.t0;

                    case 13:
                    case 'end':
                        return _context.stop();
                }
            }
        }, _callee, this, [[5, 10]]);
    }));

    return function initProjection(_x) {
        return _ref.apply(this, arguments);
    };
}();

var getReadOnlyCollectionInterface = function () {
    var _ref2 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2(repository, collectionName) {
        var collection, inf, exceptWriteFunction;
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
            while (1) {
                switch (_context2.prev = _context2.next) {
                    case 0:
                        collection = (0, _collection2.default)(repository, collectionName);
                        inf = {
                            findOne: collection.findOne,
                            find: collection.find,
                            count: collection.count
                        };

                        exceptWriteFunction = function exceptWriteFunction(funcName, collectionName) {
                            throw new Error('The ' + collectionName + ' collection\u2019s ' + funcName + ' method is not allowed on the read side');
                        };

                        return _context2.abrupt('return', Object.freeze(['insert', 'update', 'remove', 'ensureIndex', 'removeIndex'].reduce(function (acc, funcName) {
                            acc[funcName] = exceptWriteFunction.bind(null, funcName, collectionName);
                            return acc;
                        }, inf)));

                    case 4:
                    case 'end':
                        return _context2.stop();
                }
            }
        }, _callee2, this);
    }));

    return function getReadOnlyCollectionInterface(_x2, _x3) {
        return _ref2.apply(this, arguments);
    };
}();

var getCollectionInterface = function () {
    var _ref3 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee3(repository, isWriteable, collectionName) {
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
            while (1) {
                switch (_context3.prev = _context3.next) {
                    case 0:
                        if (!isWriteable) {
                            _context3.next = 2;
                            break;
                        }

                        return _context3.abrupt('return', (0, _collection2.default)(repository, collectionName));

                    case 2:
                        return _context3.abrupt('return', getReadOnlyCollectionInterface(repository, collectionName));

                    case 3:
                    case 'end':
                        return _context3.stop();
                }
            }
        }, _callee3, this);
    }));

    return function getCollectionInterface(_x4, _x5, _x6) {
        return _ref3.apply(this, arguments);
    };
}();

var listCollections = function () {
    var _ref4 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee4(repository) {
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
            while (1) {
                switch (_context4.prev = _context4.next) {
                    case 0:
                        throw new Error('todo implement me!!!');

                    case 1:
                    case 'end':
                        return _context4.stop();
                }
            }
        }, _callee4, this);
    }));

    return function listCollections(_x7) {
        return _ref4.apply(this, arguments);
    };
}();

exports.default = init;

var _collection = require('./collection');

var _collection2 = _interopRequireDefault(_collection);

var _metaCollection = require('./metaCollection');

var _metaCollection2 = _interopRequireDefault(_metaCollection);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function getStoreInterface(repository, isWriteable) {
    return Object.freeze({
        collection: getCollectionInterface.bind(null, repository, isWriteable),
        listCollections: listCollections.bind(null, repository)
    });
}

function init(repository) {
    var _this = this;

    if (repository.interfaceMap) {
        throw new Error('The read model storage is already initialized');
    }
    repository.lastTimestamp = 0;

    if (typeof repository.initHandler !== 'function') {
        repository.initHandler = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee5() {
            return regeneratorRuntime.wrap(function _callee5$(_context5) {
                while (1) {
                    switch (_context5.prev = _context5.next) {
                        case 0:
                        case 'end':
                            return _context5.stop();
                    }
                }
            }, _callee5, _this);
        }));
    }

    repository.connectionPromise = repository.connectDatabase().then(function (client) {
        return repository.client = client;
    }); // todo:
    // .then(syncronizeDatabase.bind(null, repository));

    repository.interfaceMap = new Map();
    repository.internalError = null;

    repository.metaCollection = (0, _metaCollection2.default)(repository);

    repository.readInterface = getStoreInterface(repository, false);
    repository.writeInterface = getStoreInterface(repository, true);

    repository.initDonePromise = initProjection(repository);

    return {
        getLastAppliedTimestamp: function () {
            var _ref6 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee6() {
                return regeneratorRuntime.wrap(function _callee6$(_context6) {
                    while (1) {
                        switch (_context6.prev = _context6.next) {
                            case 0:
                                _context6.next = 2;
                                return repository.connectionPromise;

                            case 2:
                                return _context6.abrupt('return', repository.lastTimestamp);

                            case 3:
                            case 'end':
                                return _context6.stop();
                        }
                    }
                }, _callee6, _this);
            }));

            return function getLastAppliedTimestamp() {
                return _ref6.apply(this, arguments);
            };
        }(),
        getReadable: function () {
            var _ref7 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee7() {
                return regeneratorRuntime.wrap(function _callee7$(_context7) {
                    while (1) {
                        switch (_context7.prev = _context7.next) {
                            case 0:
                                _context7.next = 2;
                                return repository.initDonePromise;

                            case 2:
                                return _context7.abrupt('return', repository.readInterface);

                            case 3:
                            case 'end':
                                return _context7.stop();
                        }
                    }
                }, _callee7, _this);
            }));

            return function getReadable() {
                return _ref7.apply(this, arguments);
            };
        }(),
        getError: function () {
            var _ref8 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee8() {
                return regeneratorRuntime.wrap(function _callee8$(_context8) {
                    while (1) {
                        switch (_context8.prev = _context8.next) {
                            case 0:
                                _context8.next = 2;
                                return repository.initDonePromise;

                            case 2:
                                return _context8.abrupt('return', repository.internalError);

                            case 3:
                            case 'end':
                                return _context8.stop();
                        }
                    }
                }, _callee8, _this);
            }));

            return function getError() {
                return _ref8.apply(this, arguments);
            };
        }()
    };
}
module.exports = exports['default'];